CREATED_BY_LABEL = ("created_by.gefyra.dev", "true")
CARGO_ENDPOINT_LABEL = "cargo_endpoint.gefyra.dev"
VERSION_LABEL = "version.gefyra.dev"
ACTIVE_KUBECONFIG_LABEL = "active_kubeconfig.gefyra.dev"
ACTIVE_KUBECONFIG_CONTEXT_LABEL = "active_context.gefyra.dev"
